//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  ++39(49) 827-7216     FAX: ++39(49) 827-7102
//     EMail:  loreti@padova.infn.it
//     WWW:    http://wwwcdf.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------
//
//     A C++ implementation of a simplified Boyer-Moore pattern
//     matching algorithm.  The pattern to look for is used to
//     build a bmPattern object, together with the information
//     about case significancy; for every string to be searched
//     a bmMatcher object is built from the string itself and
//     the bmPattern.  The bmPattern methods are an iterator,
//     nextMatch(), and a match-counter (nMatches()) that uses
//     nextMatch() to scan the text string.
//
// -------------------------------------------------------------
//
//     $Id: boyermoore.cxx,v 1.5 2002/12/16 10:54:03 loreti Exp $
// 
// -------------------------------------------------------------

#include <cstring>
#include <cctype>

#ifdef TEST
#include <iostream>
#include <string>
#endif // TEST

#include "boyermoore.h"

// Local functions: string copy to all lowercase characters

static void strlcpy(char *, const char *);

static void strlcpy (
  char       *dest,
  const char *source
) {
  char c;

  do {
    // SAS-C 'tolower' is a macro with side effects; couldn't write
    // while (*dest++ = tolower(*source++)) ;    :-(

    c = tolower(static_cast<unsigned char>(*source));
    *dest++ = c;
    source++;
  } while (c);
}

//      BM_TABLE_LEN is the length of the Boyer-Moore step table;
//      here is initialised to 256 to include the whole character set.

const int bmPattern::BM_TABLE_LEN = 256;

// The bmPattern class constructor.  bmPat a pointer to the original
// pattern string (or, if the case is not significant, a pointer to a
// lowercase copy of the original pattern string); bmPatLen is its
// length; and bmPos is the Boyer-Moore table holding the last position,
// in the pattern, of every character.

bmPattern::bmPattern (
  const char *pattern,
  const int   ignoreCase
) throw(std::bad_alloc)
  : bmCase(ignoreCase)
{
  bmPatLen = std::strlen(pattern);

  if (bmCase) {
    bmPat = new char[bmPatLen + 1];
    strlcpy(bmPat, pattern);
  } else {
    bmPat = (char *) pattern;
  }

  bmPos = new short[BM_TABLE_LEN];
  std::memset(bmPos, 0, BM_TABLE_LEN * sizeof(short));
  for (size_t i = 0;   i < bmPatLen;   i++) {
    bmPos[ bmPat[i] ] = i;
  }
}

// The copy constructor

bmPattern::bmPattern (
  const bmPattern &source
) throw(std::bad_alloc)
  : bmCase(source.bmCase), bmPatLen(source.bmPatLen)
{
  if (bmCase) {
    bmPat = new char[bmPatLen + 1];
    std::strcpy(bmPat, source.bmPat);
  } else {
    bmPat = source.bmPat;
  }

  bmPos = new short[BM_TABLE_LEN];
  std::memcpy(bmPos, source.bmPos, BM_TABLE_LEN * sizeof(short));
}

// The destructor

bmPattern::~bmPattern (
) throw()
{
  if (bmCase) delete [] bmPat;
  delete [] bmPos;
}

// The assignment operator

bmPattern & bmPattern::operator = (
  const bmPattern &source
) throw(std::bad_alloc)
{
  if (this != &source) {
    if (bmCase) delete [] bmPat;

    bmPatLen = source.bmPatLen;
    bmCase   = source.bmCase;

    if (bmCase) {
      bmPat = new char[bmPatLen + 1];
      std::strcpy(bmPat, source.bmPat);
    } else {
      bmPat = source.bmPat;
    }

    std::memcpy(bmPos, source.bmPos, BM_TABLE_LEN * sizeof(short));
  }

  return *this;
}

// The bmMatcher class constructor; stores a reference to the
// pattern's bmPattern object, and a pointer to the original
// string (or to a lowercase copy of the string if the case is
// to be ignored).  'pos' is the last scanned string character,
// initialised to -1; and lastPos is the position, in the string,
// of the last character to be compared.

bmMatcher::bmMatcher (
  const bmPattern &pattern,
  const char      *t
) throw(std::bad_alloc)
  : pat(pattern), pos(-1)
{
  lastPos = std::strlen(t);

  if (pat.bmCase) {
    text = new char[lastPos + 1];
    strlcpy(text, t);
  } else {
    text = (char *) t;
  }

  lastPos -= pat.bmPatLen;
}

// The destructor

bmMatcher::~bmMatcher (
) throw()
{
  if (pat.bmCase) delete [] text;
}

// nMatches returns the number of matches AFTER the character
// indexed by 'pos'; 'pos' is not modified.

int bmMatcher::nMatches (
) throw()
{
  int nMatch  = 0;
  int savePos = pos;

  while (nextMatch() >= 0) nMatch++;

  pos = savePos;
  return nMatch;
}

// nextMatch searches for the next occurrence of the pattern
// (and modifies 'pos'); the return value is the index of the
// next occurrence, or -1 for no match.

int bmMatcher::nextMatch (
) throw()
{
  pos++;

  while (pos <= static_cast<int>(lastPos)) {
    int j = pat.bmPatLen - 1;

    while (j >= 0   &&   text[pos + j]  ==  pat.bmPat[j])    j--;
    if (j < 0) return pos;

    int step = j - pat.bmPos[text[pos + j]];
    if (step > 1) pos += step;
        else      pos++;
  }

  return -1;
}

#ifdef TEST

// _Quick_and_dirty_ test program

int main()
{
  int i;
  char c;
  std::string buffer, pattern;

  for (;;) {
    std::cout << "String (or \\ to quit) ? ";
    std::cout.flush();
    std::getline(std::cin, buffer, '\n');
    if (buffer == "\\") return 0;

    std::cout << "Pattern ? ";
    std::cout.flush();
    std::getline(std::cin, pattern, '\n');

    std::cout << "Ignore case (1 for YES; 0 for NO) ? ";
    std::cout.flush();
    std::cin >> i;
    std::cin.get(c);            // Eat newline

    std::cout << "String: \"" << buffer << "\"; pattern: \"" << pattern
              << "\"; case IS ";
    if (i) std::cout << "NOT ";
    std::cout << "significant.\n";

    bmPattern bp(pattern.c_str(), i);
    bmMatcher bm(bp, buffer.c_str());
    std::cout << "NMatches: " << bm.nMatches() << std::endl;
    while ((i = bm.nextMatch()) >= 0) {
      std::cout << "Match at string[" << i << "]\n";
    }
  }
}

#endif // TEST
